;; basic drawing setup: define units and initial scale
;;____________________________________________________
;;
;; 2017-12-28 knut hohenberg (except stated otherwise)
;;

;;
;; taken  from vle-extensions.lsp (BricsCAD lisp extension compatibility layer)

(if (not vle-entget)
  (defun vle-entget ( dxf ename / res )
    (if (not (setq res (cdr (assoc dxf (entget ename)))))
      (cond  ;; retrieve omitted default values
        ((= dxf 6)   (setq res (cdr _vle:dxf_6_default)))   ;; linetype
        ((= dxf 38)  (setq res (cdr _vle:dxf_38_default)))  ;; elevation
        ((= dxf 39)  (setq res (cdr _vle:dxf_39_default)))  ;; thickness
        ((= dxf 48)  (setq res (cdr _vle:dxf_48_default)))  ;; linetype scale
        ((= dxf 60)  (setq res (cdr _vle:dxf_60_default)))  ;; visibility 0=visible
        ((= dxf 62)  (setq res (cdr _vle:dxf_62_default)))  ;; color 256=ByLayer
        ((= dxf 67)  (setq res (cdr _vle:dxf_67_default)))  ;; modelspace / paperspace entity 0=modelspace
        ((= dxf 370) (setq res (cdr _vle:dxf_370_default))) ;; lineweight -1=ByLayer
        ((= dxf 440) (setq res (cdr _vle:dxf_440_default))) ;; transparency 0=ByLayer
      )
    )
    res
  )
)

(if (not vle-list-union)
 (defun vle-list-union ( lst1 lst2 / item result )
  (foreach item lst1
   (if (not (member item result))
    (setq result (cons item result))
   )
  )
  (foreach item lst2
   (if (not (member item result))
    (setq result (cons item result))
   )
  )
  (reverse result)
 )
)

;;; from khad_lib

(defun KHAD:find_reactor (type name / rlist n reactor)
 (setq rlist (car (vlr-reactors type))
       n 1
 )
 (if rlist 
  (while n
   (setq reactor (nth n rlist))
   (if reactor
    (if (= name (vlr-data reactor))
     (setq n nil)
     (setq n (+ 1 n))
    )
    (setq n nil)
   )
  )
 )
 reactor
)

;; check for annotational text or dimension styles in the drawing

(defun KHAD:list_table (table_name xref / table_list table_item )
 (if (setq table_item (cdr (assoc 2 (tblnext table_name T))))
  (progn
   (setq table_list (list table_item))
   (while (setq table_item (tblnext table_name))
    (if xref
     (setq table_list (cons (cdr (assoc 2 table_item)) table_list))
     (if (/= 16 (logand 16 (cdr (assoc 70 table_item))))
      (setq table_list (cons (cdr (assoc 2 table_item)) table_list))
     )
    )
   )
   (reverse table_list)
  )
 )
)

(defun KHAD:has_dwg_anno_styles ( / anno tables m table table_list n item entdat xdata)
 (setq anno NIL)
 (setq tables (list "STYLE" "DIMSTYLE"))
 (setq m 0)
 (while (and m (not anno) (setq table (nth m tables)))
  (setq table_list (KHAD:list_table table nil)) ; should we check for externally referenced styles as well?
  (setq n 0)
  (while (and n (setq item (nth n table_list)))
   (if (and (not (wcmatch item "*|*,*$*")) (setq item (tblobjname table item)))
    (progn
     (setq entdat (entget item '("AcadAnnotative")))
     (if (and (setq xdata (cadr (assoc -3 entdat)))(/= (cdr (nth 4 xdata)) 0))
      (setq anno T
            n  NIL
      )
      (setq n (1+ n))
     )
    )
    (setq n (1+ n))
   )
  )
  (setq m (1+ m))
 )
 (eval anno)
)


(defun KHAD:purge_scales (/ alst cnt p elst slst)
 (setq alst (vle-list-union (KHAD:get_object_scales) (KHAD:get_mview_scales))
       cnt 0
 )
 (foreach p (dictsearch (namedobjdict) "ACAD_SCALELIST")
  (and
   (= 350 (car p))
   (setq sclnme (cdr (assoc 300 (entget (cdr p)))))
   (/= "Unscaled" sclnme)
   (not (wcmatch sclnme "*XREF*"))
   (null (member sclnme alst))
   (entdel (cdr p))
   (setq cnt (1+ cnt))
  )
 )
 (if (/= cnt 0)
  (princ (strcat "\n" (itoa cnt) " scale(s) purged."))
 )
 (princ)
)


;;check for a scale entry
(defun KHAD:check_scale (sclnme / scllst n scl check)
 (setq scllst (dictsearch (namedobjdict) "ACAD_SCALELIST")
       n 0
       check nil
 )
 (while (and n (setq scl (nth n scllst)))
  (if (and (= 350 (car scl)) (= sclnme (vle-entget 300 (cdr scl))))
   (setq check T
         n nil
   )
   (setq n (1+ n))
  )
 )
 check
)

;;create a default scale entry and make it current
(defun KHAD:add_default_scale ( / )
 ;(foreach scl (dictsearch (namedobjdict) "ACAD_SCALELIST")
 ; (if (and (= 350 (car scl)) (= "Unscaled" (vle-entget 300 (cdr scl))))
 ;  (setq defaultscale T)
 ; )
 ;)
 ;(if (not defaultscale)
 (if (not (KHAD:check_scale "Unscaled"))
  (command "_.-ScaleListEdit"
           "_Add"
           "Unscaled"
           "1:1"
           "Exit"
  )
 )
 (setvar "CANNOSCALE" "Unscaled")
)

;; rename a scale
;; append "(n)" until there are no name conflicts
(defun KHAD:rename_scale (sclnme / scllst scl sclent scldat n m orgnme oldnme)
 (foreach scl (setq scllst (dictsearch (namedobjdict) "ACAD_SCALELIST"))
  (if (and (= 350 (car scl)) (= sclnme (vle-entget 300 (cdr scl))))
   (setq sclent (cdr scl))
  )
 )
 (if sclent
  (progn
   (setq n 1
        orgnme sclnme
        oldnme sclnme
        sclnme (strcat sclnme "(" (itoa n) ")" )
   )
   (while (/= oldnme sclnme)
    (setq oldnme sclnme)
    (foreach m scllst
     (if (and (= 350 (car m)) (= sclnme (vle-entget 300 (cdr m))))
      (setq n (+ n 1)
            sclnme (strcat orgnme "(" (itoa n) ")" )
      )
     )
    )
   )
   (setq scldat (entget sclent))
   (setq scldat (subst (cons 300 sclnme) (assoc 300 scldat) scldat))
   (entmod scldat)
   sclnme
  )
 )
)



;; return a list with dotted pairs of scalenames and scalevalues
;; thanks to a forum post by Roy Klein-Gebbink
(defun KHAD:get_scales ()
 (vl-remove
  nil
  (mapcar
  '(lambda (sub)
    (if (= 350 (car sub))
     (cons
      (vle-entget 300 (cdr sub))
      (/ (vle-entget 140 (cdr sub)) (vle-entget 141 (cdr sub)))
     )
    )
   )
   (entget (vle-dictobjname (namedobjdict) "ACAD_SCALELIST"))
  )
 )
)

; return a list of scalenames used by annotational objects
(defun KHAD:get_object_scales ( / n scllst selset scl)
 (setq n 0
       scllst '()
 )
 (setq selset (ssget "X" '((0 . "*TEXT,*DIMENSION,INSERT,HATCH,MULTILEADER"))))
 (repeat (sslength selset)
  (if (setq scl (LM:annoscales (ssname selset n)))
   (setq scllst (vle-list-union scllst scl))
  )
  (setq n (1+ n))
 )
 (reverse scllst)
)

;; Annotative Scales  -  Lee Mac
;; Returns a list of all annotative scales associated with a supplied entity

(defun LM:annoscales ( ent / dic enx itm rtn )
    (if
        (and
            (setq enx (entget ent))
            (setq dic (cdr (assoc 360 (cdr (member '(102 . "{ACAD_XDICTIONARY") enx)))))
            (setq dic (cdr (assoc -1  (dictsearch dic "acdbcontextdatamanager"))))
            (setq dic (cdr (assoc -1  (dictsearch dic "acdb_annotationscales"))))
        )
        (while (setq itm (dictnext dic (not itm)))
            (setq rtn (cons (cdr (assoc 300 (entget (cdr (assoc 340 itm))))) rtn))
        )
    )
    (reverse rtn)
)

; return a list of scalenames used by mviews
(defun KHAD:get_mview_scales ( / n mviews scl scllst)
 (setq n 0
       scllst '()
 )
 (if (not (setq mviews (ssget "_X" '((0 . "VIEWPORT"))))) ; we are in modelspace, and paperspace has not been initialized
  (progn
   (setvar "TILEMODE" 0)
   (setvar "TILEMODE" 1)
   (setq mviews (ssget "_X" '((0 . "VIEWPORT"))))
  )
 )
 (repeat (sslength mviews)
  (if
   (and
    (setq scl (vle-entget 360 (ssname mviews n)))
    (setq scl (vle-entget 360 scl))
    (setq scl (vle-entget 340 scl))
    (setq scl (vle-entget 300 scl))
   )
   (if (not (member scl scllst))
    (setq scllst (cons scl scllst))
   )
  )
  (setq n (1+ n))
 )
 (reverse scllst)
)

;; round to powers of 10
(defun KHAD:round_pow_10 (num / str)
 (setq str (rtos num 1 6))
 (if (<= 1 num)
  (setq num (atoi (substr str (1+ (vl-string-search "+" str)))))
  (setq num (atoi (substr str (1+ (vl-string-search "-" str)))))
 )
 (expt 10.0 num)
)

;; return decimal places
(defun KHAD:get_decimals (num / str)
 (setq str (rtos num 1 6))
 (if (<= 1 num)
  0
  (abs (atoi (substr str (1+ (vl-string-search "-" str)))))
 )
)


;; model units converted to meters - only common AEC units handled
(defun KHAD:get_model_unit_meters (units / )
 (cond
  ((= 1 units) 0.0254)
  ((= 2 units) 0.3048)
  ((= 3 units) 1609.344)
  ((= 4 units) 0.001)
  ((= 5 units) 0.01)
  ((= 6 units) 1.0)
  ((= 7 units) 1000.0)
  (t 0)
 )
)

;; layout units converted to meters - only printer units (inch and mm) handled
(defun KHAD:get_paper_unit_meters (model_units / )
 (cond
  ((< model_units 4) 0.0254)
  ((< 3 model_units 8) 0.001)
  (t 0)
 )
)


;; populate the dialog
(defun KHAD:display_settings (units scaling scale ltscaling rwltscaling / sclval list1 settings_list)
 (setq sclval
  (if (and (/= 0 units) (/= 0 scale) (/= 0 (KHAD:get_paper_unit_meters units)))
   (if (= 0 scaling)
    (/ 1 (/ (KHAD:get_model_unit_meters units) (KHAD:get_paper_unit_meters units) scale))
    (/ (KHAD:get_model_unit_meters units) (KHAD:get_paper_unit_meters units) scale)
   )
   0
  )
 )
 ;(princ (strcat "\nScale: " (rtos scale) ", Scale value: " (rtos sclval)))
 (if (/= 0 scale) (set_tile "scale" (rtos scale)))
 (setq list1
  (list
   (if (and (= 0 scaling) (= 1 ltscaling))
    (strcat "LTSCALE:\t[" (rtos sclval 2 4) "]\tLinetypescale")
    "LTSCALE:\t[1]\tLinetypescale"
   )
   (if (or (= 0 scaling) (= 0 ltscaling))
    "MSLTSCALE:\t[0]\tUsing LTSCALE for Linetype scaling"
    "MSLTSCALE:\t[1]\tUsing CANNOSCALEVALUE for Linetype scaling"
   )
   (if (= 0 scaling)
    "PSLTSCALE:\t[0]\tMview linetypescale is model-related"
    "PSLTSCALE:\t[1]\tMview linetypescale is layout-related"
   )
   (if (= 0 scaling)
    (strcat "DIMSCALE:\t[" (rtos sclval 2 4) "]\tScale factor for dimensions")
    (strcat "CANNOSCALEVALUE:\t[" (rtos sclval 2 4) "]\tScale factor for annotational entities")
   )
   (if (= 0 scaling)
    (strcat "MLEADERSCALE:\t[" (rtos sclval 2 4) "]\tScale factor for mleaders")
    "MLEADERSCALE:\t[1]\tScale factor for mleaders"
   )
  )
 )
 (cond
  ((= units 0)
   (mode_tile "scale" 1)
   (set_tile "scale" "")
   ;(mode_tile "settings" 1)
   ;(mode_tile "scaling" 1)
   (if forced (mode_tile "accept" 1))
   (setq settings_list (list "No drafting setup possible when model units are unspecified"))
  )
  ((and (< units 4) (/= scale 0))
   (setq settings_list
    (append
     (list
      "Variable\tValue\tDescription"
      "MEASUREMENT:\t[0]\tImperial/ANSI hatches and linetypes"
      "LUNITS:\t[5]\tFractional display of units"
      "LWUNITS:\t[0]\tLineweights display in inches"
     )
     list1
     (if (and (< 1 units) (getvar "PROPUNITS")) (list (strcat "PROPUNITS:\t[" (itoa (* 2 (/ (getvar "PROPUNITS") 2))) "]\tUnits formatting")))
     '(" " "Hints:" "- Layout units assumed to be inches (SETTINGS ARE INCORRECT OTHERWISE)")
     '("- Style definitions (dimensions, mleaders) assumed to be in inches")
     (if (= 1 rwltscaling) '("- Linetypes named ms-mm_* or ms-in_* will be scaled to real-world units"))
     (if (< 1 units)
      (if (getvar "PROPUNITS")
       '("- Model and layout units differ: Length units formatting suppressed to avoid confusion"
         "- Copy/paste in paperspace will not work correctly between drawings with differing units"
         "  (the commands CC=copyclip and PC=pasteclip are supplied as a workaround)"
        )
       '("- Copy/paste in paperspace will not work correctly between drawings with differing units"
         "  (the commands CC=copyclip and PC=pasteclip are supplied as a workaround)"
       )
      )
     )
    )
   )
   (if (getvar "CANNOSCALE") (mode_tile "scaling" 0))
   (mode_tile "scale" 0)
  )
  ((and (< 3 units 8) (/= scale 0))
   (setq settings_list
    (append
     (list
      "Variable\tValue\tDescription"
      "MEASUREMENT:\t[1]\tMetric/ISO hatches and linetypes" 
      "LUNITS:\t[2]\tDecimal display of units"
      (strcat "LUPREC:\t[" (itoa (KHAD:get_decimals (* (/ 1 (KHAD:get_model_unit_meters units)) 0.000002 scale))) "]\tLinear units precision")
      "LWUNITS:\t[1]\tLineweights display in millimeters"
     )
     list1
     (if (and (< 4 units) (getvar "PROPUNITS")) (list (strcat "PROPUNITS:\t[" (itoa (* 2 (/ (getvar "PROPUNITS") 2))) "]\tUnits formatting")))
     (list (strcat "GRIDUNIT:\t[" (rtos (KHAD:round_pow_10 (* (/ 1 (KHAD:get_model_unit_meters units)) 0.002 scale)) 2 4) "]\tGrid spacing"))
     (list (strcat "SNAPUNIT:\t[" (rtos (KHAD:round_pow_10 (* (/ 1 (KHAD:get_model_unit_meters units)) 0.0002 scale)) 2 4)"]\tSnap spacing"))
     '(" " "Hints:""- Layout units assumed to be millimeters (SETTINGS ARE INCORRECT OTHERWISE)")
     '("- Style definitions (dimensions, mleaders) assumed to be in millimeters")
     (if (= 1 rwltscaling) '("- Linetypes named ms-mm_* or ms-in_* will be scaled to real-world units"))
     (if (< 4 units)
      (if (getvar "PROPUNITS")
       '("- Model and layout units differ: Length units formatting suppressed to avoid confusion"
         "- Copy/paste in paperspace will not work correctly between drawings with differing units"
         "  (the commands CC=copyclip and PC=pasteclip are supplied as a workaround)"
        )
       '("- Copy/paste in paperspace will not work correctly between drawings with differing units"
         "  (the commands CC=copyclip and PC=pasteclip are supplied as a workaround)"
       )
      )
     )
    )
   )
   (if (getvar "CANNOSCALE") (mode_tile "scaling" 0))
   (mode_tile "scale" 0)
  )
  ((< 7 units)
   (mode_tile "scale" 1)
   ;(mode_tile "settings" 1)
   (mode_tile "scaling" 1)
   (setq settings_list (list "Unsupported model units for drafting setup" "(Supported units: Inches, Feet, Miles, Millimeters, Centimeters, Meters, Kilometers)"))
  )
  ((= scale 0)
   (mode_tile "scale" 0)
   (if forced (mode_tile "accept" 1))
   (setq settings_list (list "No drafting setup possible without defined scale"))
  )
 )
 (if (getvar "CANNOSCALE")
  (setq settings_list (append settings_list '("- Unused scales will be deleted, default entry (Unscaled) will be added if not found")))
 )
 (start_list "settings")
 (mapcar 'add_list settings_list)
 (end_list)
)


;; callback functions
(defun KHAD:set_units (val / )
 (setq units (atoi val))
 (KHAD:display_settings units (atoi (get_tile "scaling")) (atof (get_tile "scale")) (atoi (get_tile "ltscaling")) (atoi (get_tile "rwltscaling"))) 
)

(defun KHAD:set_scaling (val / )
 (setq scaling (atoi val))
 (KHAD:display_settings (atoi (get_tile "model_units")) scaling (atof (get_tile "scale")) (atoi (get_tile "ltscaling")) (atoi (get_tile "rwltscaling"))) 
)

(defun KHAD:set_ltscaling (val / )
 (setq ltscaling (atoi val))
 (KHAD:display_settings (atoi (get_tile "model_units")) (atoi (get_tile "scaling")) (atof (get_tile "scale")) ltscaling (atoi (get_tile "rwltscaling"))) 
)

(defun KHAD:set_rwltscaling (val / )
 (setq rwltscaling (atoi val))
 (KHAD:display_settings (atoi (get_tile "model_units")) (atoi (get_tile "scaling")) (atof (get_tile "scale")) (atoi (get_tile "ltscaling")) rwltscaling) 
)

(defun KHAD:set_scale (val / )
 (if (or (= "" val) (< 0 (atof val)))
  (progn
   ;(set_tile "error" "")
   (setq scale (atof val))
   (if forced (mode_tile "accept" 0))
  )
  (progn
   ;(set_tile "error" "Invalid Value for Scale")
   (set_tile "scale" "")
   (mode_tile "scale" 2)
  )
 )
 (KHAD:display_settings (atoi (get_tile "model_units")) (atoi (get_tile "scaling")) scale (atoi (get_tile "ltscaling")) (atoi (get_tile "rwltscaling")))
)

(defun KHAD:chk_scale ()
 (if (< 0 (atof (get_tile "scale")))
  (done_dialog 1)
  (progn
   ;(set_tile "error" "Invalid Value for Scale")
   (set_tile "scale" "")
   (mode_tile "scale" 2)
  )
 )
)

;; main
(defun KHAD:aec_init (forced / dcl_id status oldzin scale scl sclval sclnme units_list units units_ratio scaling ltscaling rwltscaling reactor curdoc grdval)
 (graphscr)
 (if (= 1 (getvar "cvport"))
  (princ "\nThis command is only available in modelspace.")
  (progn
   (setq oldzin (getvar "DIMZIN")
   )
   (setvar "DIMZIN" 8)
   (setvar "CMDECHO" 0)
   (setq units_list
    (list
     "UNSPECIFIED"
     "Inches" "Feet" "Miles"
     "Millimeters" "Centimeters" "Meters" "Kilometers"
     "(Microinches)" "(Mils)" "(Yards)" "(Angstroms)"
     "(Nanometers)" "(Microns)" "(Decimeters)" "(Decameters)"
     "(Hectometers)" "(Gigameters)" "(Astronomical)" "(Lightyears)" "(Parsecs)"
    )
   )
   (setq units (getvar "INSUNITS"))
   (setq ltscaling 1)
   (if (and (setq reactor (KHAD:find_reactor :vlr-sysvar-reactor "KHAD:scalechange")) (vlr-added-p reactor))
    (setq rwltscaling 1)
    (setq rwltscaling 0)
   )
   (setq scaling_list '("Legacy (DIMSCALE/LTSCALE)" "Annotational (CANNOSCALE)"))
   (setq units_ratio 
    (if (/= 0 (KHAD:get_paper_unit_meters (getvar "INSUNITS")))
     (/ (KHAD:get_model_unit_meters (getvar "INSUNITS")) (KHAD:get_paper_unit_meters (getvar "INSUNITS")))
     0
    )
   )
   ;(princ (strcat "\n1.) units ratio: " (rtos units_ratio)))
   (if (KHAD:has_dwg_anno_styles)
    (setq scaling 1)
    (setq scaling 0)
   )
   (setq sclval (getvar "CANNOSCALEVALUE"))
   (if (or (/= 1 scaling) (not sclval))
    (setq sclval (getvar "DIMSCALE")
          scale (* sclval units_ratio)
    )
    (if (/= sclval 0)
     (setq scale (/ units_ratio sclval))
     (setq scale 0)
    )
   )
   ;(princ (strcat "\n2.) Scale: " (rtos scale) ", Annoscale: " (rtos sclval)))
   (setq dcl_id (load_dialog "aec_init.dcl"))
   (if (not (new_dialog "aec_init" dcl_id )) (exit))
   (if forced
    (action_tile "accept" "(chk_scale)")
    (action_tile "accept" "(done_dialog 1)")
   )
   (set_tile "title" "openAEC drafting setup")
   (start_list "model_units")
   (mapcar 'add_list units_list)
   (end_list)
   (start_list "scaling")
   (mapcar 'add_list scaling_list)
   (end_list)
   (set_tile "model_units" (itoa (getvar "INSUNITS")))
   (if scaling
    (set_tile "scaling" (itoa scaling))
    (mode_tile "scaling" 1)
   )
   (set_tile "ltscaling" (itoa ltscaling))
   (set_tile "rwltscaling" (itoa rwltscaling))
   ;(if (/= 0 scale) (set_tile "scale" (rtos scale)))
   (KHAD:display_settings (getvar "INSUNITS") scaling scale ltscaling rwltscaling)
   (setq status (start_dialog))
   (unload_dialog dcl_id)
   (if (and (= 1 status) (/= 0 units))
    (progn
     (setq curdoc (vla-get-ActiveDocument (vlax-get-acad-object)))
     (vla-EndUndoMark curdoc)
     (vla-StartUndoMark curdoc)
     (if reactor (KHAD:disable_scalechange_reactor)) ; prevent repeated firing
     (if (getvar "CANNOSCALE")
      (progn
       (KHAD:purge_scales)
       (KHAD:add_default_scale)
      )
     )
     (setvar "INSUNITS" units)
     (if (< 0 units 8)
      (progn
       (setq sclval (/ (KHAD:get_model_unit_meters units) (KHAD:get_paper_unit_meters units) scale))
       (if (< units 4)
        (progn
         (setvar "MEASUREMENT" 0)
         (setvar "LUNITS" 5)
         (setvar "LWUNITS" 0)
        )
        (progn
         (setvar "MEASUREMENT" 1)
         (setvar "LUNITS" 2)
         (setvar "LUPREC" (KHAD:get_decimals (* (/ 1 (KHAD:get_model_unit_meters units)) 0.000002 scale)))
         (setvar "LWUNITS" 1)
         (setq grdval (KHAD:round_pow_10 (* (/ 1 (KHAD:get_model_unit_meters units)) 0.002 scale)))
         (setvar "GRIDUNIT" (list grdval grdval))
         (setvar "SNAPUNIT" (list (/ grdval 10) (/ grdval 10)))
        )
       )
       (if (not (or (= 1 units) (= 4 units)))
        (if (getvar "PROPUNITS") ; this is a BricsCAD feature that is misleading when model and layout units differ
         (setvar "PROPUNITS" (* 2 (/ (getvar "PROPUNITS") 2)))
        )
       )
       (if (= 0 scaling)
        (progn
         (setvar "DIMSCALE" (/ 1 sclval))
         (command "_.-DIMSTYLE" "_AN" "_NO" (getvar "DIMSTYLE") "_YES" "_RESTORE" (getvar "DIMSTYLE"))
         (if (= 1 ltscaling)
          (progn
           (setvar "LTSCALE" (/ 1 sclval))
           (setvar "PSLTSCALE" 0)
           (setvar "MSLTSCALE" 0)
          )
          (progn
           (setvar "LTSCALE" 1)
           (setvar "PSLTSCALE" 0)
           (setvar "MSLTSCALE" 0)
          )
         )
        )
        (progn
         (command "_.-DIMSTYLE" "_AN" "_YES" (getvar "DIMSTYLE") "_YES" "_RESTORE" (getvar "DIMSTYLE"))
         (if (= 1 ltscaling)
          (progn
           (setvar "LTSCALE" 1)
           (setvar "MSLTSCALE" 1)
           (setvar "PSLTSCALE" 1)
          )
          (progn
           (setvar "LTSCALE" 1)
           (setvar "MSLTSCALE" 0)
           (setvar "PSLTSCALE" 0)
          )
         )
         (if (setq scl (assoc (setq sclnme (strcat "1:" (rtos scale))) (KHAD:get_scales)))
          (if (/= sclval (cdr scl))
           (alert (strcat "Scale " sclnme " already exists with differing scale value\n and is referenced by entities -\nManual editing required!"))
           ;(KHAD:rename_scale scl) ;this would be nice, but renaming a referenced style crashes bcad
          )
          (command "_.-ScaleListEdit"
                   "_Add"
                   sclnme
                   (strcat (rtos (/ (KHAD:get_model_unit_meters units) (KHAD:get_paper_unit_meters units))) ":" (rtos scale)) 
                   "Exit"
          )
         )
         (setvar "CANNOSCALE" sclnme)
        )
       )
       (if (= 1 rwltscaling)
        (progn
         (KHAD:realworld_ltscale)
         (KHAD:enable_scalechange_reactor)
        )
        (KHAD:disable_scalechange_reactor)
       )
      )
     )
     (vla-EndUndoMark curdoc)
    )
   )
   (setvar "DIMZIN" oldzin)
  )
 )
 (princ)
)

(defun C:AECINIT ()
 (KHAD:aec_init NIL)
)


; Copy objects to the clipboard
; adjust or disable automatic scaling if invoked in layout space
(defun KHAD:copyclip ( / curdoc units selset bspnt)
 (setq curdoc (vla-get-ActiveDocument (vlax-get-acad-object)))
 (vla-EndUndoMark curdoc)
 (vla-StartUndoMark curdoc)
 (setq units (getvar "INSUNITS"))
 (if (= 1 (getvar "CVPORT")) ; if in layout space, guess layout units
  (cond
   ((< 0 units 4)
    (setvar "INSUNITS" 1) ;we assume inches as layout-units for imperial setups (inches, feet and miles)
   )
   ((< 3 units 8)
    (setvar "INSUNITS" 4) ;we assume millimeters as layout-units for metric setups (mm, cm, m, km)
   )
   (t
    (setvar "INSUNITS" 0) ;we disable automatic scaling for all other cases (only common AEC-units handled)
   )
  )
 )
 (if (not (setq selset (ssget "I")))
  (setq selset (ssget))
 )
 (if selset
  (progn
   ;(if (< 0 (setq dynmod (getvar "DYNMODE")))
   ; (setvar "DYNMODE" (* -1 dynmod))
   ; (setq dynmod nil)
   ;)
   (initget "W U B")
   (setq bspnt (getpoint "\nSelect base point or [World origin/Ucs origin/insBase] <Ucs origin>: "))
   (cond
    ((= bspnt "W")
      (command "_.COPYBASE" "_none" (trans '(0 0 0) 1 0) selset "")
    )
    ((if (or (not bspnt) (= "U" bspnt))
      (setq bspnt '(0 0 0))
     )
     (command "_.COPYBASE" "_none" bspnt selset "")
    )
    ((= bspnt "B")
     (command "_.COPYBASE" "_none" (getvar "INSBASE") selset "")
    )
    (t
     (command "_.COPYBASE" "_none" bspnt selset "")
    )
   )
  )
 )
 (setvar "INSUNITS" units)
 (vla-EndUndoMark curdoc)
 (princ)
)


; paste objects from the clipboard
; adjust or disable automatic scaling if invoked in layout space
; create a selection from pasted entities
(defun KHAD:pasteclip ( / curdoc units lastob inspnt selset)
 (setq curdoc (vla-get-ActiveDocument (vlax-get-acad-object)))
 (vla-EndUndoMark curdoc)
 (vla-StartUndoMark curdoc)
 (setq units (getvar "INSUNITS"))
 (setq lastob (entlast))
 (if (= 1 (getvar "CVPORT")) ; if in layout space, guess layout units
  (cond
   ((< 0 units 4)
    (setvar "INSUNITS" 1)
   )
   ((< 3 units 8)
    (setvar "INSUNITS" 4)
   )
   (t
    (setvar "INSUNITS" 0)
   )
  )
 )
 (initget "W U B I")
 (setq inspnt (getpoint "\nSelect Insertion point or [World origin/Ucs origin/insBase/Interactive placement] <Ucs origin>: "))
 (cond
  ((= inspnt "W")
    (command "_.PASTECLIP" "_none" (trans '(0 0 0) 0 1) )
  )
  ((or (not inspnt) (= "U" inspnt))
    (command "_.PASTECLIP" "_none" '(0 0 0))
  )
  ((= bspnt "B")
   (command "_.PASTECLIP" "_none" (trans (getvar "INSBASE") 0 1) )
  )
  ((= "I" inspnt)
   (command "_.PASTECLIP" "_none" pause)
  )
  (t
   (command "_.PASTECLIP" "_none" inspnt)
  )
 )
 (setvar "INSUNITS" units)
 (setq selset (ssadd))
 (while (setq lastob (entnext lastob))
  (ssadd lastob selset)
 )
 (command "_.SELGRIPS" selset "")
 (vla-EndUndoMark curdoc)
 (princ)
)



(defun C:CC ()
 (KHAD:copyclip)
)

(defun C:PC ()
 (KHAD:pasteclip)
)

;;; functions to deal with real-world linetypes

(defun KHAD:enable_scalechange_reactor ( / reactor)
 (if (setq reactor (KHAD:find_reactor :vlr-sysvar-reactor "KHAD:scalechange"))
  (if (not (vlr-added-p reactor)) (vlr-add reactor))
  (setq reactor (vlr-sysvar-reactor "KHAD:scalechange" '((:vlr-sysVarChanged . KHAD:scale_callback))))
 )
 reactor
)

(defun KHAD:disable_scalechange_reactor ( / reactor)
 (if (and (setq reactor (KHAD:find_reactor :vlr-sysvar-reactor "KHAD:scalechange")) (vlr-added-p reactor))
  (vlr-remove reactor)
 )
 reactor
)

(defun KHAD:scale_callback (reactor callback_data / )
 (if
  (and
   (cadr callback_data) ;the variable was altered
   (or
    (= (car callback_data) "MSLTSCALE")
    (= (car callback_data) "LTSCALE")
    ;(= (car callback_data) "PSLTSCALE") ; real-world-scaled linetypes make little sense in paperspace
    (and (= (car callback_data) "CANNOSCALE") (= 1 (getvar "MSLTSCALE")))
   )
  )
  (progn
   (print (car callback_data))
   (KHAD:realworld_ltscale)
  )
 )
 (princ)
)

(defun KHAD:apply_ltscale (selset ltscl / n entnme entdat)
 (repeat (setq n (sslength selset))
  (setq entnme (ssname selset (setq n (1- n))))
  (setq entdat (entget entnme))
  (if (assoc 48 entdat)
   (setq entdat (subst (cons 48 ltscl) (assoc 48 entdat) entdat))
   (setq entdat (append (list (cons 48 ltscl)) entdat))
  )
  (entmod entdat)
  (entupd entnme)
 )
 (sslength selset)
)

(defun KHAD:apply_ltscale_to_block (blknme lt_mm scl_mm lt_in scl_in / entnme entdat count)
 (if (= "*" (substr blknme 1 1))
  (setq entnme nil)
  (setq entnme (cdr (assoc -2 (tblsearch "block" blknme))))
 )
 (setq count 0)
 (while (and entnme (setq entnme (entnext entnme)))
  (setq entdat (entget entnme))
  (cond
   ((and (assoc 6 entdat) (wcmatch (cdr (assoc 6 entdat)) lt_mm))
    (if (assoc 48 entdat)
     (setq entdat (subst (cons 48 scl_mm) (assoc 48 entdat) entdat)
           count (1+ count)
     )
     (setq entdat (append (list (cons 48 scl_mm)) entdat)
           count (1+ count)
     )
    )
    (entmod entdat)
    (entupd entnme)
   )
   ((and (assoc 6 entdat) (wcmatch (cdr (assoc 6 entdat)) lt_in))
    (if (assoc 48 entdat)
     (setq entdat (subst (cons 48 scl_in) (assoc 48 entdat) entdat)
           count (1+ count)
     )
     (setq entdat (append (list (cons 48 scl_in)) entdat)
           count (1+ count)
     )
    )
    (entmod entdat)
    (entupd entnme)
   )
  )
 )
 count
)

(defun KHAD:realworld_ltscale ( / scale scl_mm scl_in selset count n skplst)
 (foreach n (KHAD:list_table "LTYPE" nil)
  (if (or (wcmatch n "ms-mm_*") (wcmatch n "ms-in_*"))
   (setq scale T)
  )
 )
 (if scale
  (progn
   (setq count 0)
   (if (= 1 (getvar "MSLTSCALE"))
    (setq scale (* (/ 1(getvar "LTSCALE")) (getvar "CANNOSCALEVALUE")))
    (setq scale (/ 1 (getvar "LTSCALE")))
   )
   (setq skplst '("_ARCHTICK" "_DOTBLANK" "_DOT" "_CLOSED" "_CLOSEDBLANK" "_CLOSEDFILLED" "_OBLIQUE"))
   (if (= 1 (getvar "MEASUREMENT"))
    (cond ; we assume a metric setting with layout units = mm
     ((= 4 (getvar "INSUNITS")) (setq scl_mm scale scl_in (* scale 25.4)))
     ((= 5 (getvar "INSUNITS")) (setq scl_mm (/ scale 10.0) scl_in (/ (* scale 25.4) 10)))
     ((= 6 (getvar "INSUNITS")) (setq scl_mm (/ scale 1000.0) scl_in (/ (* scale 25.4) 1000)))
     ((= 7 (getvar "INSUNITS")) (setq scl_mm (/ scale 1000000.0) scl_in (/ (* scale 25.4) 1000000)))
     (t (setq scale nil))
    )
    (cond ; we assume layout units = in
     ((= 1 (getvar "INSUNITS")) (setq scl_in scale scl_mm (/ scale 25.4)))
     ((= 2 (getvar "INSUNITS")) (setq scl_in (/ scale 12.0) scl_mm (/ scale 12.0 25.4)))
     ((= 3 (getvar "INSUNITS")) (setq scl_in (/ scale 63360.0) scl_mm (/ scale 63360.0 25.4)))
     (t (setq scale nil))
    )
   )
   (if (and scale (setq selset (ssget "X" '((6 . "ms-mm_*")))))
    (setq count (KHAD:apply_ltscale selset scl_mm))
   )
   (if (and scale (setq selset (ssget "X" '((6 . "ms-in_*")))))
    (setq count (+ count (KHAD:apply_ltscale selset scl_in)))
   )
   (foreach n (KHAD:list_table "BLOCK" nil)
    (if (not (member n skplst))
     (setq count (+ count (KHAD:apply_ltscale_to_block n "ms-mm_*" scl_mm "ms-in_*" scl_in)))
    )
   )
   (if scale
    (princ (strcat "\nApplied real-world linetype scaling to " (itoa count) " objects."))
    (princ (strcat "\nUnsupported insunits for real-world linetype scaling."))
   )
  )
  (princ "\nNo real-world linetype objects found.")
 )
 (princ)
)

(if (not (KHAD:check_scale "Unscaled"))
 (KHAD:aec_init NIL)
)

(if (and (KHAD:find_reactor :vlr-sysvar-reactor "KHAD:scalechange") (vlr-added-p (KHAD:find_reactor :vlr-sysvar-reactor "KHAD:scalechange")))
 (progn
  (princ "\nActivating real-world linetype scaling...")
  (KHAD:realworld_ltscale)
 )
)

(princ)
